package cam.symfun;
import java.io.*;
//
//######################################################################
//                     Class ExpressionTransform
//######################################################################
//   
// Created from exressionTransform  -> Migrated C++ code to Java 3/30/98
// Chris Andersion (C) UCLA
//
//######################################################################
//
//######################################################################
//
//
public class ExpressionTransform
{

   int variableCount;
   int constantCount;
  
   int[]         executionArray;
   int       executionArraySize;
   int       evaluationDataSize;
    
   int symbolCount;
   String []  sNames;
   
   OperatorLib opLib;

   static int DELIM  = -2;
   static int VAR    = -1;
   static int NOOP   =  0;
   static int LEFTP  = -1;
   static int COMMA  =  0; 
   static int RIGHTP =  1;
   
public ExpressionTransform()
{
   variableCount      = 0;
   constantCount      = 0;
   symbolCount        = 0;
   opLib              = new RealOperatorLib(); 
   
   executionArray     = null;
   executionArraySize = 0;
   evaluationDataSize = 0;
}

public int getEvaluationDataSize()
{   
    return evaluationDataSize; 
}

public int [] getExecutionArray()
{
    return executionArray;
}
 
public String [] getSymbolicNames()
{
    return sNames;
}
//
//######################################################################
//                      MAIN
//######################################################################
//
public static void main(String[] args)
{
   

    ExpressionTransform T = new ExpressionTransform(); 
    int i;


    String[] V = new String[1];                 
    V[0]       = new String("x");
    
    String[] C = null; // new String[1];                 
    //C[0]       = new String("a");

    String E = "5.0";

    try{T.createTransform(V,C,E);}
    catch(Exception e){System.out.println(e);};
   

    System.out.println(" ");
    System.out.println(" Program End : Hit Return to Continue ");
    try{System.in.read();}catch(Exception e){};
    
}
//
//##################################################################
//  		        CREATE TRANSFORM
//##################################################################
//
public int createTransform
(String [] V, String [] C, String expressionString) throws Exception
{
    int  i;  int j;  int k;
    String ErrorString;
//
//  Capture input data
//
    variableCount   = 0;
    if(V != null) variableCount   = V.length;
    
    constantCount   = 0;
    if(C != null) constantCount   = C.length;
    
    String [] vNames = V;
	String [] cNames = C;
	
	char [] inputS = new char[expressionString.length()+1];
    expressionString.getChars(0,expressionString.length(),inputS,0);
//
//######################################################################
//
//   SEPARATE EXPRESSION INTO TOKENS
//
//######################################################################
//
    char[] S              = new char[2*strlen(inputS,0) + 1];
    S[2*strlen(inputS,0)]  = '\0';

	int tokenCount;
	tokenCount      = separateIntoTokens(inputS,S);
	if(tokenCount  <= 0){return 1;}
//
//######################################################################
//
//  Create an array of names for variables,
//  symbolic constants, and numeric constants
//
//  The indices of these symbols match the corresponding
//  index of the data array used to store their values.
//######################################################################
//
//  Count numerical constants
//
	 int numericCount = 0;
	 int index;

	 index = 0;
	 for(k = 0; k < tokenCount; k++)
	 {
	 if(((((int)(S[index]) >= 48)&&((int)(S[index]) <= 57)))
	 ||(S[index] == '.'))
	 {numericCount++;}
	 index += strlen(S,index) + 1;
	 }
//
//######################################################################
//
//  Create an array of names for variables,
//  symbolic constants, and numeric constants
//
//  The indices of these symbols match the corresponding
//  index of the data array used to store their values.
//######################################################################
//
//  Count numerical constants
//
    symbolCount   = variableCount + constantCount + numericCount;
    
	sNames = new String[symbolCount];

    for(i = 0; i < variableCount; i++)
    {
      sNames[i] = new String(vNames[i]);
	}

    for(i = variableCount, j = 0; i < variableCount + constantCount; j++,i++)
    {
      sNames[i] = new String(cNames[j]);
    }

    i = variableCount + constantCount;

    index = 0;
	for(k = 0; k < tokenCount; k++)
	{
	if(((((int)(S[index]) >= 48)&&((int)(S[index]) <= 57)))
	  ||(S[index] == '.'))
	 {
	  sNames[i] = new String(S,index,strlen(S,index));
	  i++;
	 }
	 index += strlen(S,index) + 1;
    }
/*
   System.out.println(variableCount + " " + constantCount + " " + expressionString);
   for(i = 0; i < variableCount; i++)
   System.out.println(vNames[i]);
   for(i = 0; i < constantCount; i++)
   System.out.println(cNames[i]);
   for(i = 0; i < symbolCount; i++)
   System.out.println(sNames[i]);
*/
//
//######################################################################
//
//   ENCODE TOKENS INTO ITERMEDIATE EXPRESSION CODE
//
//######################################################################
//   
    int[] expressionCode     = new int[2*tokenCount]; 
    int[] expressionSizeArg  = new int[1];
    int   expressionCodeSize;

    int encodeReturn;

    encodeReturn = encodeExpression(S,tokenCount,sNames,
    variableCount, constantCount,symbolCount,expressionCode,expressionSizeArg);
    if(encodeReturn != 0){return encodeReturn;}
    
    expressionCodeSize = expressionSizeArg[0];
//
//  code to check encoding : output should match input string
//
/*
 	for(i = 0; i < expressionCodeSize; i=i+2)
    {
    if(expressionCode[i] == VAR)
    { System.out.print(expressionCode[i+1] +  "  "); }
    else if(expressionCode[i] == DELIM)
    {
    if(expressionCode[i+1] == LEFTP)  System.out.print("(");
    if(expressionCode[i+1] == RIGHTP) System.out.print(")");
    if(expressionCode[i+1] == COMMA)  System.out.print( ",");
    }
    else if(expressionCode[i] > 0)
    {
    System.out.print(opLib.getOperatorSymbol(expressionCode[i+1]));
    System.out.print(expressionCode[i+1] +  "  "); 
    }}
    System.out.println();
    System.out.println();
*/
//
//
//######################################################################
//
//   ERROR CHECKING
//
//######################################################################
//
    int parenthesisCheck = 0;
	for(i = 0; i < expressionCodeSize; i+=2)
    {
    if(expressionCode[i] == DELIM) parenthesisCheck  += expressionCode[i+1];
    }
    if(parenthesisCheck != 0)
    {
    ErrorString = 
    "Error            = Unbalanced Paranthesis  " + "\n" + 
    "Offending String = " + expressionString + "\n";
    Exception e = new Exception(ErrorString); 
    throw e;
    }
//
//######################################################################
//
//   CREATE EXECUTION CODE : PACK INTO EXECUTIONARRAY
//
//######################################################################
//
//
//  Count operators and arguments
//
//  Create storage space for executionArray
//
    int opCount  = 0;
    int argCount = 0;
	 for(i=0; i < expressionCodeSize; i+=2)
    {
    if(expressionCode[i] > 0)
    {
    opCount++;
    argCount +=
    opLib.getOperatorArgCount(expressionCode[i+1]) + 1;
    }}

    executionArraySize = 2*opCount + argCount;
    executionArray     = new int[executionArraySize];

    int  executionIndex = 0;
    int  dataIndex      = symbolCount - 1;
    
    int[] executionIndexArg = new int[1];
    int[] dataIndexArg      = new int[1];
//
//  Treat expressions with just one symbol 
//  (Modification 2/26/97)
//
    if(executionArraySize == 0)
    {
     executionArraySize = 4;
     executionArray     = new int[executionArraySize];
     executionArray[0]  = 0;  // set operand to unary +
     executionArray[1]  = 2;  // two arguments
     executionArray[2]  = expressionCode[1];
     executionArray[3]  = symbolCount;
     evaluationDataSize = symbolCount + 1;
     return 0;
    }
//
//
//
    int   istart;
	int   iend;
//
//  Scan for paranthesis depth
//
	int pDepth    = 0;
    for(i=0; i < expressionCodeSize; i+=2)
    {
    if((expressionCode[i] == DELIM)&&(expressionCode[i+1]== LEFTP)) pDepth++;
    }
//
//  Setup up execution sequences for each depth of paranthesis
//
    int pIndex;  int jp;
    int sReturn;
    for(k = pDepth; k >= 1; k--)
    {
     pIndex = 0;
     for(j = 0; j < expressionCodeSize; j+=2)
     {
     if((expressionCode[j] == DELIM)&&(expressionCode[j+1]== LEFTP)) pIndex++;

     if(pIndex == k)
     {
     expressionCode[j] = NOOP; expressionCode[j+1] = NOOP;  // consume (
     istart = j;
     jp = j+2;
     while(!((expressionCode[jp] == DELIM)&&(expressionCode[jp+1] == RIGHTP)))
     {jp+=2;}
     iend   = jp;
     expressionCode[jp] = NOOP;expressionCode[jp+1] = NOOP; // consume )
     
     executionIndexArg[0] = executionIndex;
     dataIndexArg[0]      = dataIndex;
     sReturn = setupEvaluation(expressionCode,istart,iend,  // process
                     executionArray,executionIndexArg, dataIndexArg);           
     executionIndex = executionIndexArg[0];
     dataIndex      = dataIndexArg[0];  
     
     if(sReturn != 0){return sReturn;}
     pIndex--;
     }}
    }
//
//  Do pDepth == 0 directly
//
    istart   = 0;
    iend     = expressionCodeSize;
    
    executionIndexArg[0] = executionIndex;
    dataIndexArg[0]      = dataIndex;
    sReturn = setupEvaluation(expressionCode,istart,iend,executionArray,
                    executionIndexArg, dataIndexArg);
    executionIndex = executionIndexArg[0];
    dataIndex      = dataIndexArg[0];
    
    if(sReturn != 0){return sReturn;}
//
//  set evaluationDataSize required
//
    evaluationDataSize = dataIndex+1;
//
//  Check expressioncode to see only one operand left
//
    int resultCount = 0;
	for(i = 0; i < expressionCodeSize; i+=2)
    {
    if(expressionCode[i] != NOOP) resultCount++;
    }
    if(resultCount != 1)
    {
    ErrorString = 
    "Error            = Illegal Expression  " + "\n" + 
    "Offending String = " + expressionString + "\n";
    Exception e = new Exception(ErrorString); 
    throw e;  
    }
//
//  Error checking ... see what the execution Array looks like..
//
/*
	System.out.println("Execution Array Size : " + executionArraySize);
	System.out.println("EvaluationDataSize   : " + evaluationDataSize);
	for(i = 0; i < executionArraySize; i++)
	System.out.print(executionArray[i] + " ");
	System.out.println();
*/
//
//######################################################################
//
//   CLEAN UP
//
//######################################################################
//
    return 0;
}
//
//##################################################################
//  		           SETUPEVALUATION
//##################################################################
//
int setupEvaluation(int [] expressionCode,
int istart, int iend, int [] executionArray, int [] executionIndexArg,
int [] dataIndexArg) throws Exception
{
//
//  This routine repeatedly scans the expressionCode from left
//  to right and sets up the exection pointers and indices.
//
//  The expressionCode is scanned from i = istart to i < iend;
//
//  The execution sequence is coded into executionArray
//  in the form :
//
//  [operatorIndex][argument Count][ ...argument indices ... ]
//
//
//  It is assumed that the input expression code contains no
//  paranthesis.
//
    int dataIndex      = dataIndexArg[0];
    int executionIndex = executionIndexArg[0];
    
    int i; int j; int k;
    int argIndex      = 0; 
    int argIndexLeft  = 0; 
    int argIndexRight = 0;
    int argCount;
    String ErrorString;
//
//  remove ,'s
//
	for(i = istart; i < iend; i=i+2)
    {
    if((expressionCode[i] == DELIM)&&(expressionCode[i+1]== COMMA))
    {expressionCode[i] = NOOP; expressionCode[i+1] = NOOP;}
    }
//
//  Function calls (priority 1)
//
	for(i = istart; i < iend; i= i+2)
    {
    if(expressionCode[i] == 1)
    {
    //
    // set function call
    //
    argCount = opLib.getOperatorArgCount(expressionCode[i+1]);
    executionArray[executionIndex] = expressionCode[i+1]; executionIndex++;
    executionArray[executionIndex] = argCount + 1;        executionIndex++;
    //
    // scan right for argCount arguments and pack indices
    //
    k = 0;
 	for(j=i+2; j < iend; j=j+2)
    {
      if(expressionCode[j] != NOOP)
      {
      if(expressionCode[j] != VAR)
      {
      ErrorString = 
      "Error            = Incorrect Number of Arguments  " + "\n" + 
      "Offending String = " + opLib.getOperatorSymbol(expressionCode[i+1]) + "\n";
      Exception e = new Exception(ErrorString); 
      throw e;  
      }
      else
      {
      argIndex            = expressionCode[j+1];   // get location
      expressionCode[j]   = NOOP;                  // consume argument
      expressionCode[j+1] = NOOP;
      executionArray[executionIndex] = argIndex; executionIndex++;
      k++;
      if(k == argCount) j = iend;
      }}
    }
    //
    // ran off end looking for arguments
    //
    if(k!= argCount)
    {
    ErrorString = 
    "Error            = Incorrect Number of Arguments  " + "\n" + 
    "Offending String = " + opLib.getOperatorSymbol(expressionCode[i+1]) + "\n";
    Exception e = new Exception(ErrorString); 
    throw e;  
    }
    //
    // set pointer to return argument
    //
    dataIndex++;
    executionArray[executionIndex] =  dataIndex;  executionIndex++;
    //
    // replace operator code by pointer to it's result
    //
    expressionCode[i]   = VAR;
    expressionCode[i+1] = dataIndex;
    }}
//
//  Exponentiation priority = 2
//
    for(i=istart; i < iend; i= i+2)
    {

    if(expressionCode[i] == 2)
    {
    //
    // set function call
    //
    executionArray[executionIndex] = expressionCode[i+1]; executionIndex++;
    executionArray[executionIndex] = 3;                   executionIndex++;
    //
    // scan right for arguments
    //
    for(j=i+2; j < iend; j=j+2)
    {
      if(expressionCode[j] != NOOP)
      {
      if(expressionCode[j] != VAR)
      {
      ErrorString = 
      "Error            = Incorrect Number of Arguments  " + "\n" + 
      "Offending String = " + opLib.getOperatorSymbol(expressionCode[i+1]) + "\n";
      Exception e = new Exception(ErrorString); 
      throw e;  
      }
      else
      {
      argIndexRight       = expressionCode[j+1];   // get location
      expressionCode[j]   = NOOP;                  // consume argument
      expressionCode[j+1] = NOOP;
      j = iend;
      }}
    }
    //
    // scan left for arguments
    //
    for(j=i-2; j >= istart; j=j-2)
    {
      if(expressionCode[j] != NOOP)
      {
      if(expressionCode[j] != VAR)
      {
      ErrorString = 
      "Error            = Incorrect Number of Arguments  " + "\n" + 
      "Offending String = " + opLib.getOperatorSymbol(expressionCode[i+1]) + "\n";
      Exception e = new Exception(ErrorString); 
      throw e;  
      }
      else
      {
      argIndexLeft        = expressionCode[j+1];   // get location
      expressionCode[j]   = NOOP;                  // consume argument
      expressionCode[j+1] = NOOP;
      j = istart;
      }}
    }
    //
    // set up evaluation pointers
    //
    dataIndex++;
    executionArray[executionIndex] = argIndexLeft;   executionIndex++;
    executionArray[executionIndex] = argIndexRight;  executionIndex++;
    executionArray[executionIndex] = dataIndex;      executionIndex++;
    //
    // replace operator code by pointer to it's result
    //
    expressionCode[i]  = VAR;
    expressionCode[i+1]= dataIndex;
    }}   
    
//
//  Unary Operators (priority 3)
//
    for(i=istart; i < iend; i= i+2)
    {

    if(expressionCode[i] == 3)
    {
    //
    // set function call
    //
    executionArray[executionIndex] = expressionCode[i+1]; executionIndex++;
    executionArray[executionIndex] = 2;                   executionIndex++;
    //
    // scan right for arguments
    //
    for(j=i+2; j < iend; j=j+2)
    {
      if(expressionCode[j] != NOOP)
      {
      if(expressionCode[j] != VAR)
      {
      ErrorString = 
      "Error            = Incorrect Number of Arguments  " + "\n" + 
      "Offending String = " + opLib.getOperatorSymbol(expressionCode[i+1]) + "\n";
      Exception e = new Exception(ErrorString); 
      throw e;  
      }
      else
      {
      argIndex = expressionCode[j+1];       // get location
      expressionCode[j]   = NOOP;           // consume argument
      expressionCode[j+1] = NOOP;
      j = iend;
      }}
    }
    //
    // set up evaluation pointers
    //
    dataIndex++;
    executionArray[executionIndex] = argIndex;  executionIndex++;
    executionArray[executionIndex] = dataIndex; executionIndex++;
    //
    // replace operator code by pointer to it's result
    //
    expressionCode[i]  = VAR;
    expressionCode[i+1]= dataIndex;
    }}
//
//  Binary Operators (priority 4,5)
//
    for(k = 4; k <= 5; k++)
    {

    for(i=istart; i < iend; i= i+2)
    {

    if(expressionCode[i] == k)
    {
    //
    // set function call
    //
    executionArray[executionIndex] = expressionCode[i+1]; executionIndex++;
    executionArray[executionIndex] = 3;                   executionIndex++;
    //
    // scan right for arguments
    //
    for(j=i+2; j < iend; j=j+2)
    {
      if(expressionCode[j] != NOOP)
      {
      if(expressionCode[j] != VAR)
      {
      ErrorString = 
      "Error            = Incorrect Number of Arguments  " + "\n" + 
      "Offending String = " + opLib.getOperatorSymbol(expressionCode[i+1]) + "\n";
      Exception e = new Exception(ErrorString); 
      throw e;  
      }
      else
      {
      argIndexRight       = expressionCode[j+1];   // get location
      expressionCode[j]   = NOOP;                  // consume argument
      expressionCode[j+1] = NOOP;
      j = iend;
      }}
    }
    //
    // scan left for arguments
    //
    for(j=i-2; j >= istart; j=j-2)
    {
      if(expressionCode[j] != NOOP)
      {
      if(expressionCode[j] != VAR)
      {
      ErrorString = 
      "Error            = Incorrect Number of Arguments  " + "\n" + 
      "Offending String = " + opLib.getOperatorSymbol(expressionCode[i+1]) + "\n";
      Exception e = new Exception(ErrorString); 
      throw e;  
      }
      else
      {
      argIndexLeft        = expressionCode[j+1];   // get location
      expressionCode[j]   = NOOP;                  // consume argument
      expressionCode[j+1] = NOOP;
      j = istart;
      }}
    }
    //
    // set up evaluation pointers
    //
    dataIndex++;
    executionArray[executionIndex] = argIndexLeft;   executionIndex++;
    executionArray[executionIndex] = argIndexRight;  executionIndex++;
    executionArray[executionIndex] = dataIndex;      executionIndex++;
    //
    // replace operator code by pointer to it's result
    //
    expressionCode[i]  = VAR;
    expressionCode[i+1]= dataIndex;
    }}

    }
    dataIndexArg[0]      = dataIndex;
    executionIndexArg[0] = executionIndex;
    return 0;
}
//
//##################################################################
//  		            ENCODE EXPRESSION
//##################################################################
//
int encodeExpression(char[] S, int tokenCount,
String [] sNames, int vCount, int cCount, int sCount,
int[] expressionCode, int[] expressionCodeSize) throws Exception
{
//
// This routine encodes the tokenized experession in S into an
// intermediate expression code. 
//
// Input parameters :
//
// S           = string in which tokens are separated by string terminators
// tokenCount  = total number of tokens in  S
//
// sNames = array of names of variables, symbolic and numeric constants
// vCount = number of variable names
// cCount = number of symbolic constant names
// sCount = total number of symbol names
//
// Return Arguments :
//
// expressionCode     = array of ints to hold code (space allocated
//					    externally)
//
// expressionCodeSize = size of expressionCode array
//

//
// Declare and initialize encoding variables
//
	int  i;
	int  k;
    int tokenIndex = 0;
    int opIndex;
    int checkFlag;
    int unaryFlag;
    String ErrorString;

    int unaryPlusIndex   = opLib.getUnaryOperatorIndex("+");
    int unaryMinusIndex  = opLib.getUnaryOperatorIndex("-");
    int binaryPlusIndex  = opLib.getBinaryOperatorIndex("+");
    int binaryMinusIndex = opLib.getBinaryOperatorIndex("-");

    int binaryTimesIndex  = opLib.getOperatorIndex("*");
    int binaryDivideIndex = opLib.getOperatorIndex("/");
    int binaryExpoIndex   = opLib.getOperatorIndex("^");


	char[] Stoken = new char[256];
	int StokenLength;
//
//  Main encoding loop
//
	 int index = 0;
	 for(k = 0; k < tokenCount; k++)
	 {
	 strcpy(Stoken,S,index);
	 StokenLength = strlen(Stoken,0);
	 
//
//   Delimiters : Coded (DELIM, Index of delimiter)
//
	if (Stoken[0] == '(')
	{
	 expressionCode[tokenIndex] = DELIM; tokenIndex++;
     expressionCode[tokenIndex] = LEFTP; tokenIndex++;
	}
	else if (Stoken[0] == ')')
	{
     expressionCode[tokenIndex] =  DELIM; tokenIndex++;
     expressionCode[tokenIndex] =  RIGHTP; tokenIndex++;
	}
	else if (Stoken[0] == ',')
	{
     expressionCode[tokenIndex] =  DELIM; tokenIndex++;
	 expressionCode[tokenIndex] =  COMMA; tokenIndex++;
	}
//
//  Numerical Constants : Coded (VAR, Index of evaluationData value)
//
    else if(((((int)(Stoken[0]) >= 48)&&((int)(Stoken[0]) <= 57)))
    ||(Stoken[0] == '.'))
    {
    for(i = cCount + vCount; i < sCount; i++)
    {
    if(strcmp(Stoken,sNames[i]) == 0)
    {
       expressionCode[tokenIndex] = VAR;  tokenIndex++;
       expressionCode[tokenIndex] =   i;  tokenIndex++;
       i = sCount;
    }}
   }
//
//  Symbolic Tokens
//
  	else if(    (((int)(Stoken[0]) >= 65)&&((int)(Stoken[0]) <= 90))
             || (((int)(Stoken[0]) >= 97)&&((int)(Stoken[0]) <= 122))
           )
    {
//
//  Variables  : Coded (VAR, Index of evaluationData value)
//
    checkFlag = 0;
    for(i = 0; i < vCount; i++)
    {
    if(strcmp(Stoken,sNames[i]) == 0)
    {
       expressionCode[tokenIndex] = VAR;  tokenIndex++;
       expressionCode[tokenIndex] =   i;  tokenIndex++;
       checkFlag = 1;
    }}
//
//  Symbolic Constants  : Coded (VAR, Index of evaluationData value)
//
    if(checkFlag == 0)
    {
    for(i = vCount; i < vCount + cCount; i++)
    {
    if(strcmp(Stoken,sNames[i]) == 0)
    {
       expressionCode[tokenIndex] = VAR;  tokenIndex++;
       expressionCode[tokenIndex] =  i;   tokenIndex++;
       checkFlag = 1;
    }
    }}
//
//  Symbolic Operators : Coded (priority value, Index of Function)
//
    if(checkFlag == 0)
    {
     opIndex = opLib.getOperatorIndex(new String(Stoken,0,StokenLength));
     if(opIndex >= 0)
     {
      expressionCode[tokenIndex] =
      opLib.getOperatorPriority(opIndex);
      tokenIndex++;
      expressionCode[tokenIndex] =  opIndex;
      tokenIndex++;
      checkFlag = 1;
    }}
    //
    // Have a symbolic constant which doesn't match operator, variable
    // or a constant
    //
    if(checkFlag != 1)
    {
    ErrorString = 
    "Error            = Illegal Symbol  " + "\n" + 
    "Offending Symbol = " + (new String(Stoken,0,StokenLength)) + "\n";
     Exception e = new Exception(ErrorString); 
     throw e;  
    }
    }
//
//  Algebraic Operators : Coded (priority value, Index of Function)
//
  	else if(Stoken[0] == '+')
    {

    if(tokenIndex == 0)                          // first operand
    { unaryFlag = 1;}
    else if
    (
    (expressionCode[tokenIndex  -2] == DELIM)&&     // delimiter
    ((expressionCode[tokenIndex -1] == LEFTP)||     //   (
    (expressionCode[tokenIndex  -1] == COMMA))      //   ,
    )
    { unaryFlag = 1;}
    else if
    (
    (expressionCode[tokenIndex  -2] > 0)&&                 // operand
    ((expressionCode[tokenIndex -1] == binaryPlusIndex)||  //   +
     (expressionCode[tokenIndex -1] == binaryMinusIndex)|| //   -
     (expressionCode[tokenIndex -1] == binaryDivideIndex)||//   /
     (expressionCode[tokenIndex -1] == binaryExpoIndex)||  //   +
     (expressionCode[tokenIndex -1] == binaryTimesIndex))  //   *
    )
    { unaryFlag = 1;}
    else
    {
    unaryFlag = 0;
    }

    if(unaryFlag == 1)
    {
      expressionCode[tokenIndex] =
      opLib.getOperatorPriority(unaryPlusIndex);
      tokenIndex++;
      expressionCode[tokenIndex] =  unaryPlusIndex;
      tokenIndex++;
    }
    else
    {
 	  expressionCode[tokenIndex] =
      opLib.getOperatorPriority(binaryPlusIndex);
      tokenIndex++;
      expressionCode[tokenIndex] =  binaryPlusIndex;
      tokenIndex++;
    }
    }
//
//
//
    else if(Stoken[0] == '-')
    {
    if(tokenIndex == 0)                          // first operand
    { unaryFlag = 1;}
    else if
    (
    (expressionCode[tokenIndex  -2] == DELIM)&&     // delimiter
    ((expressionCode[tokenIndex -1] == LEFTP)||     //   (
    (expressionCode[tokenIndex  -1] == COMMA))      //   ,
    )
    { unaryFlag = 1;}
    else if
    (
    (expressionCode[tokenIndex  -2] > 0)&&                 // operand
    ((expressionCode[tokenIndex -1] == binaryPlusIndex)||  //   +
     (expressionCode[tokenIndex -1] == binaryMinusIndex)|| //   -
     (expressionCode[tokenIndex -1] == binaryDivideIndex)||//   /
     (expressionCode[tokenIndex -1] == binaryExpoIndex)||  //   +
     (expressionCode[tokenIndex -1] == binaryTimesIndex))  //   *
    )
    { unaryFlag = 1;}
    else
    {
    unaryFlag = 0;
    }

    if(unaryFlag == 1)
    {
      expressionCode[tokenIndex] =
      opLib.getOperatorPriority(unaryMinusIndex);
      tokenIndex++;
      expressionCode[tokenIndex] =  unaryMinusIndex;
      tokenIndex++;
    }
    else
    {
 	  expressionCode[tokenIndex] =
      opLib.getOperatorPriority(binaryMinusIndex);
      tokenIndex++;
      expressionCode[tokenIndex] =  binaryMinusIndex;
      tokenIndex++;
    }
//
//                     *, /, ^
//
    }
    else if((Stoken[0] == '*')||(Stoken[0] == '/')||(Stoken[0] == '^'))
    {
     opIndex = opLib.getOperatorIndex(new String(Stoken,0,StokenLength));
     expressionCode[tokenIndex] =
     opLib.getOperatorPriority(opIndex);
     tokenIndex++;
     expressionCode[tokenIndex] =  opIndex;
     tokenIndex++;
    }
    else
    {
    ErrorString = 
    "Error            = Illegal Symbol  " + "\n" + 
    "Offending Symbol = " + (new String(Stoken,0,StokenLength)) + "\n";
     Exception e = new Exception(ErrorString); 
     throw e; 
    }
 //
 // update token pointer
 //
    index += strlen(S,index) + 1;
    }

    expressionCodeSize[0] = tokenIndex;
    return 0;
 }
//
//##################################################################
//  		        SEPARATE_INTO_TOKENS
//##################################################################
//
public int separateIntoTokens(char []sIn, char []S) 
{
//
//  This routine takes the string sIn and decomposes it into
//  distinct tokens separated by spaces '\0'. The storage for
//  S must be declared externally (maximal size will be
//  (2*strlen(sIn) + 1).  No error checking is done to see if S
//  is of sufficient size.
//
//  The value returned is the number of tokens ...
//
    char [] sInput  = new char[strlen(sIn,0) + 1];  //  capture input string
    strcpy(sInput, sIn);                        //
//   
    int i;
//
//  Remove Spaces
//
    int  iMark = 0;
    for(i = 0; i < strlen(sInput,0); i++)
    {
    if(sInput[i] != ' '){ sInput[iMark] = sInput[i]; iMark++; }
    }
    sInput[iMark] = '\0';
//
//  Insert Spaces
//
    iMark = 0;
    for(i = 0; i < strlen(sInput,0); i++)
    {
//
//  Operands and delimiters
//
	 if( (((int)(sInput[i]) >= 40)&&((int)(sInput[i]) <= 45))
        ||(sInput[i] == '/')
        ||(sInput[i] == '^')
		) {S[iMark] = sInput[i]; iMark++; S[iMark] = ' '; iMark++;}

    else
//
//  Variables and symbolic constants
//
    if(    (((int)(sInput[i]) >= 65)&&((int)(sInput[i]) <= 90))
        || (((int)(sInput[i]) >= 97)&&((int)(sInput[i]) <= 122))
      )
    {
    while
      (    (((int)(sInput[i]) >= 65)&&((int)(sInput[i]) <= 90))
        || (((int)(sInput[i]) >= 97)&&((int)(sInput[i]) <= 122))
        || (((int)(sInput[i]) >= 48)&&((int)(sInput[i]) <= 57))
      ){ S[iMark] = sInput[i]; iMark++; i++; }
      i--;
		S[iMark] = ' '; iMark++;
    }
//
//  Digits
//
    else
    if (((((int)(sInput[i]) >= 48)&&((int)(sInput[i]) <= 57)))
      ||(sInput[i] == '.'))
    {
     while(((int)(sInput[i]) >= 48)&&((int)(sInput[i]) <= 57))
     {S[iMark] = sInput[i]; iMark++; i++; }
 //
 //  Decimal Numbers
 //
     if(sInput[i] == '.')
     {
     S[iMark] = '.'; iMark++; i++;
     while(((int)(sInput[i]) >= 48)&&((int)(sInput[i]) <= 57))
     {S[iMark] = sInput[i]; iMark++; i++; }
     }
 //
 //  Exponential Notation
 //
     if((sInput[i] == 'e')||(sInput[i] == 'E'))
     {
     S[iMark] = 'e'; iMark++; i++;
     if(sInput[i] == '+'){S[iMark] = '+'; iMark++; i++;}
     if(sInput[i] == '-'){S[iMark] = '-'; iMark++; i++;}
     while(((int)(sInput[i]) >= 48)&&((int)(sInput[i]) <= 57))
     {S[iMark] = sInput[i]; iMark++; i++; }
     }
//
//  Insert Space
//
     i--;
	  S[iMark] = ' '; iMark++;

    }
    else
    {
    System.out.println(" Error : Illegal Symbol  ");
    System.out.println(" Offending String :  "+ sIn);
    System.out.println(" Offending Symbol :  "+ sInput[i]);
	return -1;
	 }

    }
//
//  Mark the end of the string so that strlen will give us the actual number
//  of S used.
//
	 S[iMark] = '\0';
//
//  Capture the size of the new string and then insert string terminators
//  between the tokens
//
    int  Ssize = strlen(S,0);
    for(i=0; i < Ssize; i++) if(S[i] == ' ') S[i] = '\0';
//
//  Count tokens
//
	 int tokenCount = 0;
	 int index      = 0;
	 while(index < Ssize)
    {
	 tokenCount++;
	 index += strlen(S,index) + 1;
	 }
	 if(tokenCount == 0){return -1;}
//
//  Clean up and then return tokenCount ...
//
	 return tokenCount;
}

//
// These functions on character arrays, are meant to 
// mimic their C counterparts. 
// 

//
// Find the length of the string starting at S[index] and terminating
// with a '/0'.
//
public int strlen(char[] S, int index)
{
    int  i      = index;
    int  cCount = 0; 
    while(S[i] != '\0'){i++; cCount++;}
    return cCount;   
}
//
// Copy the string to sA from sB
//
public void strcpy(char[] sA, char [] sB)
{
    int i = 0;
    while(sB[i] != '\0'){sA[i]=sB[i]; i++;}
    sA[i] = '\0';
}

public void strcpy(char[] sA, char [] sB, int index)
{
    int i = index;
    int j = 0;
    while(sB[i] != '\0'){sA[j]=sB[i]; i++; j++;}
    sA[j] = '\0';
}

public int strcmp(char [] S1, String S2)
{
    int length = strlen(S1,0);
    String S = new String(S1,0,length);
    return S.compareTo(S2);
}

}