package cam.symfun;
//
//######################################################################
//                     Class RealOperatorLib
//######################################################################
//   
// Created from CAMrealOperatorLib -> Migrated C++ code to Java 3/30/98
// Chris Andersion (C) UCLA
//
//######################################################################
//
//######################################################################
//
//
import java.Math.*;

public class RealOperatorLib implements OperatorLib
{

public RealOperatorLib(){};
//
// Static Initialization
//
public static int operatorCount = 26;

public static String [] Symbols   =
{"+", "-", "+", "-", "*", "/","^","sin", "cos",
"tan","asin","acos","atan","atan2",
"sinh","cosh","tanh",
"ceil","exp","fabs","floor","fmod","log","log10","sqrt","pow"};

public static int [] Priority =
{3,3,5,5,4,4,2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};

public static int [] ArgCount  =
{1,1,2,2,2,2,2,1,1,1,1,1,1,2,1,1,1,1,1,1,1,2,1,1,1,2};

public void evaluate(int fcnIndex, int [] argIndex, double [] evaluationData)
{
switch(fcnIndex)
{
case(0) : // Plus
{evaluationData[argIndex[1]] 
=  +(evaluationData[argIndex[0]]);};
break;

case(1) : // Minus
{evaluationData[argIndex[1]] 
=  -(evaluationData[argIndex[0]]);};
break;

case(2) : // Add
{evaluationData[argIndex[2]] 
=  (evaluationData[argIndex[0]])+(evaluationData[argIndex[1]]);};
break;

case(3) : // Subtract
{evaluationData[argIndex[2]] 
=  (evaluationData[argIndex[0]])-(evaluationData[argIndex[1]]);};
break;

case(4) : // Times
{evaluationData[argIndex[2]] 
= (evaluationData[argIndex[0]])*(evaluationData[argIndex[1]]);};
break;


case(5) : // Divide
{evaluationData[argIndex[2]] 
=  (evaluationData[argIndex[0]])/(evaluationData[argIndex[1]]);};
break;

case(6) : // Exponentiate
{evaluationData[argIndex[2]] 
=  Math.pow(evaluationData[argIndex[0]],evaluationData[argIndex[1]]);};
break;

case(7) : // Sin
{evaluationData[argIndex[1]] 
= Math.sin(evaluationData[argIndex[0]]);};
break;

case(8) : // Cos
{evaluationData[argIndex[1]] 
= Math.cos(evaluationData[argIndex[0]]);};
break;

case(9) : // Tan
{evaluationData[argIndex[1]] 
= Math.tan(evaluationData[argIndex[0]]);};
break;

case(10) : // Asin
{evaluationData[argIndex[1]] 
= Math.asin(evaluationData[argIndex[0]]);};
break;

case(11) : // Acos
{evaluationData[argIndex[1]] 
= Math.acos(evaluationData[argIndex[0]]);};
break;

case(12) : // Atan
{ evaluationData[argIndex[1]] 
= Math.atan(evaluationData[argIndex[0]]);};
break;

case(13) : // Atan2
{evaluationData[argIndex[2]] 
= Math.atan2(evaluationData[argIndex[0]],evaluationData[argIndex[1]]);};
break;

case(14) : // Sinh
{	
double x = evaluationData[argIndex[0]]; 
evaluationData[argIndex[1]] 
= (Math.exp(x) - Math.exp(-x))/2.0;};
break;

case(15) : // Cosh
{	
double x = evaluationData[argIndex[0]]; 
evaluationData[argIndex[1]] 
= (Math.exp(x) + Math.exp(-x))/2.0;};
break;

case(16) : // Tanh
{	
double x = evaluationData[argIndex[0]]; 
evaluationData[argIndex[1]] 
= (Math.exp(x) - Math.exp(-x))/(Math.exp(x) + Math.exp(-x));};
break;

case(17) : // Ceil
{evaluationData[argIndex[1]] 
= Math.ceil(evaluationData[argIndex[0]]);};
break;

case(18) : // Exp
{evaluationData[argIndex[1]] 
= Math.exp(evaluationData[argIndex[0]]);};
break;

case(19) : // Fabs
{evaluationData[argIndex[1]] 
= Math.abs(evaluationData[argIndex[0]]);};
break;

case(20) : // Floor
{evaluationData[argIndex[1]] 
= Math.floor(evaluationData[argIndex[0]]);};
break;

case(21) : // Fmod
{evaluationData[argIndex[2]] 
= Math.IEEEremainder(evaluationData[argIndex[0]],evaluationData[argIndex[1]]);};
break;

case(22) : // Log
{evaluationData[argIndex[1]] 
= Math.log(evaluationData[argIndex[0]]);};
break;

case(23) : // Log10
{evaluationData[argIndex[1]] 
= Math.log(evaluationData[argIndex[0]])/Math.log(10.0);};
break;

case(24) : // Sqrt
{evaluationData[argIndex[1]] 
= Math.sqrt(evaluationData[argIndex[0]]);};
break;

case(25) : // Pow
{evaluationData[argIndex[2]] 
= Math.pow(evaluationData[argIndex[0]],evaluationData[argIndex[1]]);};
break;
} // end case
}

public int getOperatorIndex(String Sym)
{
    int ireturn = -1;
    int i;
	for(i=0; i< operatorCount; i++)
    {
	if(Sym.equals(Symbols[i])) ireturn = i;
    }
    return ireturn;
}

public int getUnaryOperatorIndex(String Sym)
{
    int ireturn = -1;
	if(Sym.equals("+")) ireturn = 0;
    if(Sym.equals("-")) ireturn = 1;
    return ireturn;
}

public int getBinaryOperatorIndex(String Sym)
{
    int ireturn = -1;
	if(Sym.equals("+")) ireturn = 2;
    if(Sym.equals("-")) ireturn = 3;
    return ireturn;
}

public int getOperatorPriority(int index)
{
 	return Priority[index];
}

public int getOperatorArgCount(int index)
{
 	return ArgCount[index];
}

public int getOperatorCount()
{
 	return operatorCount;
}

public String getOperatorSymbol(int index)
{
 	return new String(Symbols[index]);
}

}